/*
 * Author:
 *   Sven Czarnian <devel@svcz.de>
 * Brief:
 *   Implements the backend receiver
 * Copyright:
 *   2021 Sven Czarnian
 * License:
 *   GNU General Public License v3 (GPLv3)
 */

#include <aman/com/BackendReceiver.h>

#include "ZmqContext.h"

using namespace aman;
using namespace std::chrono;

BackendReceiver::BackendReceiver() noexcept :
        m_socket(),
        m_receiverThread(),
        m_stopReceiver(false),
        m_sequences(),
        m_sequencesLock() { }

bool BackendReceiver::initialize(const Communication& configuration) {
    if (nullptr != this->m_socket)
        return true;
    if (false == configuration.valid)
        return false;

    this->m_socket = std::make_unique<zmq::socket_t>(ZmqContext::instance().context(), zmq::socket_type::sub);

    /* configure the encryption */
    if (false == this->setSocketKey(configuration.serverPublicIdentifier, zmq::sockopt::curve_serverkey))
        return false;
    if (false == this->setSocketKey(configuration.clientPublicIdentifier, zmq::sockopt::curve_publickey))
        return false;
    if (false == this->setSocketKey(configuration.clientPrivateIdentifier, zmq::sockopt::curve_secretkey))
        return false;

    /* connect to the server */
    try {
        this->m_socket->connect("tcp://" + configuration.address + ":" + std::to_string(configuration.portNotification));
    }
    catch (zmq::error_t&) {
        this->m_socket = std::unique_ptr<zmq::socket_t>();
        return false;
    }

    this->m_stopReceiver = false;
    this->m_receiverThread = std::thread(&BackendReceiver::run, this);

    return true;
}

bool BackendReceiver::deinitialize() {
    if (nullptr == this->m_socket)
        return true;

    this->m_stopReceiver = true;
    this->m_receiverThread.join();
    this->m_socket->close();
    this->m_socket = std::make_unique<zmq::socket_t>();

    return true;
}

bool BackendReceiver::initialized() const noexcept {
    return nullptr != this->m_socket;
}

void BackendReceiver::receiveSequence() {
    zmq::message_t message;

    if (nullptr == this->m_socket)
        return;

    try {
        auto result = this->m_socket->recv(message, zmq::recv_flags::dontwait);
        if (false == result.has_value() || 0 == result.value())
            return;
    }
    catch (zmq::error_t&) {
        return;
    }

    std::unique_ptr<aman::AircraftSequence> retval = std::make_unique<aman::AircraftSequence>();
    retval->ParseFromString(reinterpret_cast<const char*>(message.data()));

    std::lock_guard guard(this->m_sequencesLock);
    this->m_sequences.push_back(std::move(retval));
}

void BackendReceiver::run() {
    while (false == this->m_stopReceiver) {
        auto message = this->receive();
        if (nullptr == message) {
            std::this_thread::sleep_for(500ms);
            continue;
        }

        this->receiveSequence();
    }
}

std::shared_ptr<aman::AircraftSequence> BackendReceiver::receive() {
    std::lock_guard guard(this->m_sequencesLock);
    if (0 != this->m_sequences.size()) {
        auto retval = this->m_sequences.front();
        this->m_sequences.pop_front();
        return retval;
    }

    return nullptr;
}

BackendReceiver& BackendReceiver::instance() noexcept {
    static BackendReceiver __instance;
    return __instance;
}
